# coding=utf8
"""
webhook.py - Sopel GitHub Module
Copyright 2015 Max Gurela
Copyright 2019 dgw

 _______ __ __   __           __
|     __|__|  |_|  |--.--.--.|  |--.
|    |  |  |   _|     |  |  ||  _  |
|_______|__|____|__|__|_____||_____|
 ________         __     __                 __
|  |  |  |.-----.|  |--.|  |--.-----.-----.|  |--.-----.
|  |  |  ||  -__||  _  ||     |  _  |  _  ||    <|__ --|
|________||_____||_____||__|__|_____|_____||__|__|_____|

"""

from __future__ import unicode_literals

from sopel import tools
from sopel.formatting import bold, color
from sopel.tools.time import get_timezone, format_time

from .formatting import get_formatted_response
from .formatting import fmt_repo
from .formatting import fmt_name

from threading import Thread
import bottle
import json
import requests

# Because I'm a horrible person
sopel_instance = None

def setup_webhook(sopel):
    global sopel_instance
    sopel_instance = sopel
    host = sopel.config.github.webhook_host
    port = sopel.config.github.webhook_port

    base = StoppableWSGIRefServer(host=host, port=port)
    server = Thread(target=bottle.run, kwargs={'server': base})
    server.setDaemon(True)
    server.start()
    sopel.memory['gh_webhook_server'] = base
    sopel.memory['gh_webhook_thread'] = server

    conn = sopel.db.connect()
    c = conn.cursor()

    try:
        c.execute('SELECT * FROM gh_hooks')
    except Exception:
        create_table(sopel, c)
        conn.commit()
    conn.close()


def create_table(bot, c):
    primary_key = '(channel, repo_name)'

    c.execute('''CREATE TABLE IF NOT EXISTS gh_hooks (
        channel TEXT,
        repo_name TEXT,
        enabled BOOL DEFAULT 1,
        url_color TINYINT DEFAULT 2,
        tag_color TINYINT DEFAULT 6,
        repo_color TINYINT DEFAULT 13,
        name_color TINYINT DEFAULT 15,
        hash_color TINYINT DEFAULT 14,
        branch_color TINYINT DEFAULT 6,
        PRIMARY KEY {0}
        )'''.format(primary_key))


def shutdown_webhook(sopel):
    global sopel_instance
    sopel_instance = None
    if 'gh_webhook_server' in sopel.memory:
        print('Stopping webhook server')
        sopel.memory['gh_webhook_server'].stop()
        sopel.memory['gh_webhook_thread'].join()
        print('GitHub webhook shutdown complete')


class StoppableWSGIRefServer(bottle.ServerAdapter):
    server = None

    def run(self, handler):
        from wsgiref.simple_server import make_server, WSGIRequestHandler
        if self.quiet:
            class QuietHandler(WSGIRequestHandler):
                def log_request(*args, **kw):
                    pass
            self.options['handler_class'] = QuietHandler
        self.server = make_server(self.host, self.port, handler, **self.options)
        self.server.serve_forever()

    def stop(self):
        self.server.shutdown()
        self.server.server_close()


def get_targets(repo):
    conn = sopel_instance.db.connect()
    c = conn.cursor()
    c.execute('SELECT * FROM gh_hooks WHERE repo_name = ? AND enabled = 1', (repo.lower(), ))
    return c.fetchall()


def process_payload(payload, targets):
    if payload['event'] == 'ping':
        for row in targets:
            sopel_instance.say('[{}] {}: {} (Your webhook is now enabled)'.format(
                          fmt_repo(payload['repository']['name'], row),
                          fmt_name(payload['sender']['login'], row),
                          payload['zen']), row[0])
        return

    for row in targets:
        messages = get_formatted_response(payload, row)
        # Write the formatted message(s) to the channel
        for message in messages:
            sopel_instance.say(message, row[0])


@bottle.get("/webhook")
def show_hook_info():
    return 'Listening for webhook connections!'


@bottle.post("/webhook")
def webhook():
    try:
        payload = bottle.request.json
    except:
        return bottle.abort(400, 'Something went wrong!')

    payload['event'] = bottle.request.headers.get('X-GitHub-Event') or 'ping'
    targets = get_targets(payload['repository']['full_name'])

    # process hook payload in background
    payload_handler = Thread(target=process_payload, args=(payload, targets))
    payload_handler.start()

    # send HTTP response ASAP, hopefully within GitHub's very short timeout
    return '{"channels":' + json.dumps([target[0] for target in targets]) + '}'


@bottle.get('/auth')
def handle_auth_response():
    code = bottle.request.query.code
    state = bottle.request.query.state

    repo = state.split(':')[0]
    channel = state.split(':')[1]

    data = {'client_id': sopel_instance.config.github.client_id,
            'client_secret': sopel_instance.config.github.client_secret,
            'code': code}
    raw = requests.post('https://github.com/login/oauth/access_token', data=data, headers={'Accept': 'application/json'})
    try:
        res = json.loads(raw.text)

        if 'error' in res:
            raise ValueError('{err}: {desc}'.format(err=res['error'], desc=res['error_description']))
        if 'scope' not in res:
            raise ValueError('You\'ve already completed authorization on this repo')
        if 'write:repo_hook' not in res['scope']:
            raise ValueError('You didn\'t allow read/write on repo hooks!')

        access_token = res['access_token']

        data = {
            "name": "web",
            "active": True,
            "events": ["*"],
            "config": {
                "url": sopel_instance.config.github.external_url,
                "content_type": "json"
            }
        }

        raw = requests.post('https://api.github.com/repos/{}/hooks?access_token={}'.format(repo, access_token), data=json.dumps(data))
        res = json.loads(raw.text)

        if 'ping_url' not in res:
            if 'errors' in res:
                raise ValueError(', '.join([error['message'] for error in res['errors']]))
            else:
                raise ValueError('Webhook creation failed, try again.')

        raw = requests.get(res['ping_url'] + '?access_token={}'.format(access_token))

        title = 'Done!'
        header = 'Webhook setup complete!'
        body = 'That was simple, right?! You should be seeing a completion message in {} any second now'.format(channel)
        flair = 'There\'s no way it was that easy… things are never this easy…'
    except Exception as e:
        title = 'Error!'
        header = 'Webhook setup failed!'
        body = 'Please try using the link in {} again, something went wrong!'.format(channel)
        flair = str(e)

    page = '''
<!DOCTYPE html>
<html>
  <head>
    <title>{title}</title>
    <style>
      body {{
        width: 35em;
        margin: 0 auto;
        font-family: Tahoma, Verdana, Arial, sans-serif;
      }}
    </style>
  </head>
  <body>
    <h1>{header}</h1>
    <p>{body}</p>
    <small><em>{flair}</em></small>
  </body>
</html>
    '''

    return page.format(title=title, header=header, body=body, flair=flair)



# coding=utf8
"""Sopel GitHub

GitHub module for Sopel
"""
from __future__ import unicode_literals, absolute_import, division, print_function

from .github import *

__author__ = 'maxpowa'
__email__ = 'maxpowa@outlook.com'
__maintainer__ = 'dgw'
__version__ = '0.4.3'
__repo__ = 'https://github.com/sopel-irc/sopel-github'




# coding=utf8
"""
github.py - Sopel GitHub Module
Copyright 2015 Max Gurela
Copyright 2019 dgw

 _______ __ __   __           __
|     __|__|  |_|  |--.--.--.|  |--.
|    |  |  |   _|     |  |  ||  _  |
|_______|__|____|__|__|_____||_____|

"""

from __future__ import unicode_literals
from sopel import tools
from sopel.module import OP, NOLIMIT, commands, example, require_chanmsg, rule, url
from sopel.formatting import bold, color, monospace
from sopel.tools.time import get_timezone, format_time
from sopel.config.types import StaticSection, ValidatedAttribute

from . import formatting
from .formatting import shorten_url, emojize
from .webhook import setup_webhook, shutdown_webhook

import operator
from collections import deque

import sys
if sys.version_info.major < 3:
    from urllib import urlencode
    from urllib2 import HTTPError
else:
    from urllib.parse import urlencode
    from urllib.error import HTTPError
import json
import requests
import re
import datetime
import base64

'''
 _______           __         __
|   |   |.-----.--|  |.--.--.|  |.-----.
|       ||  _  |  _  ||  |  ||  ||  -__|
|__|_|__||_____|_____||_____||__||_____|

'''

# GitHub enforces alphanumeric usernames, and allows only one punctuation character: hyphen ('-')
# Regex copied and slightly modified to meet our needs from CC0 source:
# https://github.com/shinnn/github-username-regex/blob/0794566cc10e8c5a0e562823f8f8e99fa044e5f4/module.js#L1
githubUsername = (
    r'(?!(?:collections|events|sponsors|topics|trending)/)'  # exclude special sections
    r'[A-Za-z\d](?:[A-Za-z\d]|-(?=[A-Za-z\d])){0,38}'
)
# GitHub additionally allows dots ('.') in repo names, as well as hyphens
# not copied from anywhere, but handy to simply reuse
githubRepoSlug = r'[A-Za-z0-9\.\-]+'
# lots of regex and other globals to make this stuff work
baseURL = r'https?://(?:www\.)?github\.com/({username}/{repo})'.format(username=githubUsername, repo=githubRepoSlug)
repoURL = baseURL + r'/?(?!\S)'
issueURL = baseURL + r'/(?:issues|pull)/([\d]+)(?:#issuecomment-([\d]+))?'
commitURL = baseURL + r'/(?:commit)/([A-z0-9\-]+)'
contentURL = baseURL + r'/(?:blob|raw)/([^/\s]+)/([^#\s]+)(?:#L(\d+)(?:-L(\d+))?)?'


class GitHubSection(StaticSection):
    client_id = ValidatedAttribute('client_id', default=None)
    client_secret = ValidatedAttribute('client_secret', default=None)
    secret = ValidatedAttribute('secret', default=None)  # TODO remove in 0.3.0 or thereabouts
    webhook = ValidatedAttribute('webhook', bool, default=False)
    webhook_host = ValidatedAttribute('webhook_host', default='0.0.0.0')
    webhook_port = ValidatedAttribute('webhook_port', default='3333')
    external_url = ValidatedAttribute('external_url', default='http://your_ip_or_domain_here:3333')


def configure(config):
    config.define_section('github', GitHubSection, validate=False)
    config.github.configure_setting('client_id', 'GitHub API Client ID')
    config.github.configure_setting('client_secret', 'GitHub API Client Secret')
    config.github.configure_setting('webhook', 'Enable webhook listener functionality')
    if config.github.webhook:
        config.github.configure_setting('webhook_host', 'Listen IP for incoming webhooks (0.0.0.0 for all IPs)')
        config.github.configure_setting('webhook_port', 'Listen port for incoming webhooks')
        config.github.configure_setting('external_url', 'Callback URL for webhook activation, should be your externally facing domain or IP. You must include the port unless you are reverse proxying.')


def setup(sopel):
    sopel.config.define_section('github', GitHubSection)

    if sopel.config.github.webhook:
        setup_webhook(sopel)

    if not sopel.config.github.client_secret:
        if sopel.config.github.secret:
            sopel.config.github.client_secret = sopel.config.github.secret
            del sopel.config.github.secret
            sopel.config.save()
            tools.stderr("[GitHub] Migrated `secret` to `client_secret` in config.")


def shutdown(sopel):
    shutdown_webhook(sopel)

'''
 _______ ______ _____        ______                    __
|   |   |   __ |     |_     |   __ |.---.-.----.-----.|__|.-----.-----.
|   |   |      <       |    |    __||  _  |   _|__ --||  ||     |  _  |
|_______|___|__|_______|    |___|   |___._|__| |_____||__||__|__|___  |
                                                                |_____|
'''


def fetch_api_endpoint(bot, url):
    # GitHub deprecated passing authentication via query parameters in November
    # 2019. Passing OAuth client credentials as user/password instead is the
    # supported replacement:
    # https://developer.github.com/changes/2020-02-10-deprecating-auth-through-query-param/
    auth = None
    if bot.config.github.client_id and bot.config.github.client_secret:
        auth = (bot.config.github.client_id, bot.config.github.client_secret)
    return requests.get(url, auth=auth).text


@rule(r'.*(?<!\S)/?#(\d+)\b.*')
@require_chanmsg
def issue_reference(bot, trigger):
    """
    Separate function to work around Sopel not loading rules/commands for @url callables.
    """
    issue_info(bot, trigger)


@url(issueURL)
def issue_info(bot, trigger, match=None):
    if match:  # Link triggered
        repo = match.group(1)
        num = match.group(2)
        comment_id = match.group(3)
        URL = 'https://api.github.com/repos/%s/issues/%s' % (repo, num)
        if comment_id:
            URL = 'https://api.github.com/repos/%s/issues/comments/%s' % (repo, comment_id)
    else:  # Issue/PR number triggered
        repo = bot.db.get_channel_value('github_issue_repo', trigger.sender)
        num = trigger.group(1)
        if not repo:
            return NOLIMIT
        URL = 'https://api.github.com/repos/%s/issues/%s' % (repo, num)

    try:
        raw = fetch_api_endpoint(bot, URL)
    except HTTPError:
        bot.say('[GitHub] API returned an error.')
        return NOLIMIT
    data = json.loads(raw)
    try:
        body = data['body']
    except (KeyError):
        bot.say('[GitHub] API says this is an invalid issue. Please report this if you know it should work!')
        return NOLIMIT

    if body is None or body.strip() == '':
        body = 'No description provided.'
    else:
        body = formatting.fmt_short_comment_body(body)

    # what we have so far
    response = [
        bold('[GitHub]'),
        ' [',
        repo,
        ' #',
        num,
        '] ',
    ]

    if comment_id:
        # comment format is simple
        response.extend([
            'Comment',
        ])
    else:
        # if it's a link directly to the issue/PR, things are more complicated
        type_ = 'issue'
        state = data['state']
        if 'pull_request' in data:
            type_ = 'PR'

        if type_ == 'PR' and state == 'closed':
            # annoying consequence of "all PRs are issues, but not all issues are PRs"
            # merge status is only included if the object is fetched via `pulls` endpoint
            try:
                pr_raw = fetch_api_endpoint(bot, data['pull_request']['url'])
            except HTTPError:
                # just use the "issue" state, fine
                pass

            pr_data = json.loads(pr_raw)
            if pr_data.get("merged"):
                state = "merged"

        response.extend([
            state,
            ' ',
            type_,
        ])

    # reunited once again; the rest of the output format is common
    response.extend([
        ' by ',
        data['user']['login'],
        ': '
    ])

    if ('title' in data):
        # (well, *almost* common)
        response.append(emojize(data['title']))
        response.append(bold(' | '))
    response.append(emojize(body))

    # append link, if not triggered by a link
    if not match:
        link = shorten_url(data['html_url'])
        response.append(bold(' | '))
        response.append(link)

    bot.say(''.join(response))


@commands('gh-repo')
@example('.gh-repo sopel-irc/sopel-github')
@require_chanmsg('[GitHub] You can only link a repository to a channel.')
def manage_channel_repo(bot, trigger):
    """
    Set the repository to use for looking up standalone issue/PR references.
    """
    allowed = bot.channels[trigger.sender].privileges.get(trigger.nick, 0) >= OP
    if not allowed and not trigger.admin:
        return bot.say('You must be a channel operator to use this command!')

    if not trigger.group(2):
        return bot.say(bot.db.get_channel_value('github_issue_repo', trigger.sender))

    bot.db.set_channel_value('github_issue_repo', trigger.sender, trigger.group(3))
    bot.reply('Set linked repo for %s to %s.' % (trigger.sender, trigger.group(3)))


@url(commitURL)
def commit_info(bot, trigger, match=None):
    match = match or trigger
    URL = 'https://api.github.com/repos/%s/commits/%s' % (match.group(1), match.group(2))

    try:
        raw = fetch_api_endpoint(bot, URL)
    except HTTPError:
        bot.say('[GitHub] API returned an error.')
        return NOLIMIT
    data = json.loads(raw)
    try:
        lines = data['commit']['message'].splitlines()
        if len(lines) > 1:
            body = lines[0] + '…'
        elif len(lines) > 0:
            body = lines[0]
        else:
            body = ''
    except (KeyError):
        bot.say('[GitHub] API says this is an invalid commit. Please report this if you know it\'s a correct link!')
        return NOLIMIT

    if body.strip() == '':
        body = 'No commit message provided.'

    file_count = len(data['files'])
    response = [
        bold('[GitHub]'),
        ' [',
        match.group(1),
        '] ',
        data['author']['login'] if data['author'] else data['commit']['author']['name'],
        ': ',
        body,
        bold(' | '),
        str(data['stats']['total']),
        ' changes in ',
        str(file_count),
        ' file' if file_count == 1 else ' files'
    ]
    bot.say(''.join(response))


@url(contentURL)
def file_info(bot, trigger, match=None):
    match = match or trigger
    repo = match.group(1)
    path = match.group(3)
    ref = match.group(2)
    start_line = match.group(4)
    end_line = match.group(5)
    URL = 'https://api.github.com/repos/%s/contents/%s?ref=%s' % (repo, path, ref)

    try:
        raw = fetch_api_endpoint(bot, URL)
    except HTTPError:
        bot.say('[GitHub] API returned an error.')
        return NOLIMIT
    data = json.loads(raw)

    if data.get('type', 'dir') != 'file':
        # silently ignore directory contents (and malformed responses) for now
        return NOLIMIT

    response = [
        bold('[GitHub]'),
        ' [',
        repo,
        '] ',
        data['path'],
        ' @ ',
        ref,
    ]

    if start_line:
        lines = base64.b64decode(data['content']).splitlines()

        try:
            snippet = lines[int(start_line) - 1].decode('utf-8')
        except (IndexError, UnicodeDecodeError):
            # Line doesn't exist, or not a text file
            snippet = None

        if snippet:
            response.extend([
                ' | L',
                start_line,
                ': ',
                monospace(snippet),
                ' […] (to L%s)' % end_line if end_line else '',
            ])

    bot.say(''.join(response))


def get_data(bot, trigger, URL):
    URL = URL.split('#')[0]
    try:
        raw = fetch_api_endpoint(bot, URL)
        rawLang = fetch_api_endpoint(bot, URL + '/languages')
    except HTTPError:
        bot.say('[GitHub] API returned an error.')
        return NOLIMIT
    data = json.loads(raw)
    langData = list(json.loads(rawLang).items())
    langData = sorted(langData, key=operator.itemgetter(1), reverse=True)

    if 'message' in data:
        return bot.say('[GitHub] %s' % data['message'])

    langColors = deque(['12', '08', '09', '13'])

    max = sum([pair[1] for pair in langData])

    data['language'] = ''
    for (key, val) in langData[:3]:
        data['language'] = data['language'] + color(str("{0:.1f}".format(float(val) / max * 100)) + '% ' + key, langColors[0]) + ' '
        langColors.rotate()

    if len(langData) > 3:
        remainder = sum([pair[1] for pair in langData[3:]])
        data['language'] = data['language'] + color(str("{0:.1f}".format(float(remainder) / max * 100)) + '% Other', langColors[0]) + ' '

    timezone = get_timezone(bot.db, bot.config, None, trigger.nick)
    if not timezone:
        timezone = 'UTC'
    data['pushed_at'] = format_time(bot.db, bot.config, timezone, trigger.nick, trigger.sender, from_utc(data['pushed_at']))

    return data


@url(repoURL)
def repo_info(bot, trigger, match=None):
    user, repo = [s.strip() for s in match.group(1).split('/', 1)]
    URL = 'https://api.github.com/repos/%s/%s' % (user, repo)
    fmt_response(bot, trigger, URL, True)


@commands('github', 'gh')
@example('.gh sopel-irc/sopel-github')
def github_repo(bot, trigger):
    repo = trigger.group(3) or None

    if repo is None:
        return bot.reply('I need a repository name, or `user/reponame`.')

    if repo.lower() == 'version':
        return bot.say('[GitHub] Version {} by {}, report issues at {}'.format(
            github.__version__, github.__author__, github.__repo__))

    if repo.lower() == 'status':
        current = json.loads(requests.get('https://status.github.com/api/status.json').text)
        lastcomm = json.loads(requests.get('https://status.github.com/api/last-message.json').text)

        status = current['status']
        if status == 'major':
            status = "\x02\x034Broken\x03\x02"
        elif status == 'minor':
            status = "\x02\x037Shakey\x03\x02"
        elif status == 'good':
            status = "\x02\x033Online\x03\x02"

        lstatus = lastcomm['status']
        if lstatus == 'major':
            lstatus = "\x02\x034Broken\x03\x02"
        elif lstatus == 'minor':
            lstatus = "\x02\x037Shakey\x03\x02"
        elif lstatus == 'good':
            lstatus = "\x02\x033Online\x03\x02"

        timezone = get_timezone(bot.db, bot.config, None, trigger.nick)
        if not timezone:
            timezone = 'UTC'
        lastcomm['created_on'] = format_time(bot.db, bot.config, timezone, trigger.nick, trigger.sender, from_utc(lastcomm['created_on']))

        return bot.say('[GitHub] Current Status: ' + status + ' | Last Message: ' + lstatus + ': ' + lastcomm['body'] + ' (' + lastcomm['created_on'] + ')')
    elif repo.lower() == 'rate-limit':
        return bot.say(fetch_api_endpoint(bot, 'https://api.github.com/rate_limit'))

    if '/' not in repo:
        repo = trigger.nick.strip() + '/' + repo
    URL = 'https://api.github.com/repos/%s' % (repo.strip())

    fmt_response(bot, trigger, URL)


def from_utc(utcTime, fmt="%Y-%m-%dT%H:%M:%SZ"):
    """
    Convert UTC time string to time.struct_time
    """
    return datetime.datetime.strptime(utcTime, fmt)


def fmt_response(bot, trigger, URL, from_regex=False):
    data = get_data(bot, trigger, URL)

    if not data:
        return

    response = [
        bold('[GitHub]'),
        ' ',
        str(data['full_name'])
    ]

    if data['description'] != None:
        response.append(' - ' + str(emojize(data['description'])))

    if not data['language'].strip() == '':
        response.extend([' | ', data['language'].strip()])

    response.extend([
        ' | Last Push: ',
        str(data['pushed_at']),
        ' | Stargazers: ',
        str(data['stargazers_count']),
        ' | Watchers: ',
        str(data['subscribers_count']),
        ' | Forks: ',
        str(data['forks_count']),
        ' | Network: ',
        str(data['network_count']),
        ' | Open Issues: ',
        str(data['open_issues'])
    ])

    if not from_regex:
        response.extend([' | ', data['html_url']])

    bot.say(''.join(response))


@commands('gh-hook')
@require_chanmsg('[GitHub] GitHub hooks can only be configured in a channel')
@example('.gh-hook maxpowa/Inumuta enable')
def configure_repo_messages(bot, trigger):
    '''
    .gh-hook <repo> [enable|disable] - Enable/disable displaying webhooks from repo in current channel (You must be a channel OP)
    Repo notation is just <user/org>/<repo>, not the whole URL.
    '''
    allowed = bot.channels[trigger.sender].privileges.get(trigger.nick, 0) >= OP
    if not allowed and not trigger.admin:
        return bot.say('You must be a channel operator to use this command!')

    if not trigger.group(2):
        return bot.say(configure_repo_messages.__doc__.strip())

    channel = trigger.sender.lower()
    repo_name = trigger.group(3).lower()

    if not '/' in repo_name or 'http://' in repo_name or 'https://' in repo_name:
        return bot.say('Invalid repo formatting, see "{}help gh-hook" for an example'.format(bot.config.core.help_prefix))

    enabled = True if not trigger.group(4) or trigger.group(4).lower() == 'enable' else False

    auth_data = {
        'client_id': bot.config.github.client_id,
        'scope': 'write:repo_hook',
        'state': '{}:{}'.format(repo_name, channel)}
    auth_url = 'https://github.com/login/oauth/authorize?{}'.format(urlencode(auth_data))

    conn = bot.db.connect()
    c = conn.cursor()

    c.execute('SELECT * FROM gh_hooks WHERE channel = ? AND repo_name = ?', (channel, repo_name))
    result = c.fetchone()
    if not result:
        c.execute('''INSERT INTO gh_hooks (channel, repo_name, enabled) VALUES (?, ?, ?)''', (channel, repo_name, enabled))
        bot.say("Successfully enabled listening for {repo}'s events in {chan}.".format(chan=channel, repo=repo_name))
        bot.say('Great! Please allow me to create my webhook by authorizing via this link: ' + shorten_url(auth_url))
        bot.say('Once that webhook is successfully created, I\'ll post a message in here. Give me about a minute or so to set it up after you authorize. You can configure the colors that I use to display webhooks with {}gh-hook-color'.format(bot.config.core.help_prefix))
    else:
        c.execute('''UPDATE gh_hooks SET enabled = ? WHERE channel = ? AND repo_name = ?''', (enabled, channel, repo_name))
        bot.say("Successfully {state} the subscription to {repo}'s events".format(state='enabled' if enabled else 'disabled', repo=repo_name))
        if enabled:
            bot.say('Great! Please allow me to create my webhook by authorizing via this link: ' + shorten_url(auth_url))
            bot.say('Once that webhook is successfully created, I\'ll post a message in here. Give me about a minute or so to set it up after you authorize. You can configure the colors that I use to display webhooks with {}gh-hook-color'.format(bot.config.core.help_prefix))
    conn.commit()
    conn.close()


@commands('gh-hook-color')
@require_chanmsg('[GitHub] GitHub hooks can only be configured in a channel')
@example('.gh-hook-color maxpowa/Inumuta 13 15 6 6 14 2')
def configure_repo_colors(bot, trigger):
    '''
    .gh-hook-color <repo> <repo color> <name color> <branch color> <tag color> <hash color> <url color> - Set custom colors for the webhook messages (Uses mIRC color indicies)
    '''
    allowed = bot.channels[trigger.sender].privileges.get(trigger.nick, 0) >= OP
    if not allowed and not trigger.admin:
        return bot.say('You must be a channel operator to use this command!')

    if not trigger.group(2):
        return bot.say(configure_repo_colors.__doc__.strip())

    channel = trigger.sender.lower()
    repo_name = trigger.group(3).lower()
    colors = []
    try:
        colors = [int(c) % 16 for c in trigger.group(2).replace(trigger.group(3), '', 1).split()]
    except:
        return bot.say('You must provide exactly 6 colors that are integers and are space separated. See "{}help gh-hook-color" for more information.'.format(bot.config.core.help_prefix))

    if len(colors) != 6:
        return bot.say('You must provide exactly 6 colors! See "{}help gh-hook-color" for more information.'.format(bot.config.core.help_prefix))

    conn = bot.db.connect()
    c = conn.cursor()

    c.execute('SELECT * FROM gh_hooks WHERE channel = ? AND repo_name = ?', (channel, repo_name))
    result = c.fetchone()
    if not result:
        return bot.say('Please use "{}gh-hook {} enable" before attempting to configure colors!'.format(bot.config.core.help_prefix, repo_name))
    else:
        combined = colors
        combined.append(channel)
        combined.append(repo_name)
        c.execute('''UPDATE gh_hooks SET repo_color = ?, name_color = ?, branch_color = ?, tag_color = ?,
                     hash_color = ?, url_color = ? WHERE channel = ? AND repo_name = ?''', combined)
        conn.commit()
        c.execute('SELECT * FROM gh_hooks WHERE channel = ? AND repo_name = ?', (channel, repo_name))
        row = c.fetchone()
        bot.say("[{}] Example name: {} tag: {} commit: {} branch: {} url: {}".format(
                formatting.fmt_repo(repo_name, row),
                formatting.fmt_name(trigger.nick, row),
                formatting.fmt_tag('tag', row),
                formatting.fmt_hash('c0mm17', row),
                formatting.fmt_branch('master', row),
                formatting.fmt_url('http://git.io/', row)))



# coding=utf8
"""
formatting.py - Sopel GitHub Module
Copyright 2015 Max Gurela
Copyright 2019 dgw

 _______ __ __   __           __
|     __|__|  |_|  |--.--.--.|  |--.
|    |  |  |   _|     |  |  ||  _  |
|_______|__|____|__|__|_____||_____|
 _______                             __   __   __
|    ___|.-----.----.--------.---.-.|  |_|  |_|__|.-----.-----.
|    ___||  _  |   _|        |  _  ||   _|   _|  ||     |  _  |
|___|    |_____|__| |__|__|__|___._||____|____|__||__|__|___  |
                                                        |_____|
"""

from __future__ import unicode_literals

import re
import textwrap

import requests

from sopel.formatting import color
from sopel import tools

try:
    import emoji
except ImportError:
    emojize = lambda text: text
else:
    emojize = lambda text: emoji.emojize(text, use_aliases=True)

current_row = None
current_payload = None

LOGGER = tools.get_logger('github')


def fmt_url(s, row=None):
    if not row:
        row = current_row
    return color(s, fg=row[3])


def fmt_tag(s, row=None):
    if not row:
        row = current_row
    return color(s, fg=row[4])


def fmt_repo(s, row=None):
    if not row:
        row = current_row
    return color(s, fg=row[5])


def fmt_name(s, row=None):
    if not row:
        row = current_row
    return color(s, fg=row[6])


def fmt_hash(s, row=None):
    if not row:
        row = current_row
    return color(s, fg=row[7])


def fmt_branch(s, row=None):
    if not row:
        row = current_row
    return color(s, fg=row[8])


def fmt_short_comment_body(body):
    lines = [
        line.strip()
        for line in body.splitlines()
        if line
        and line[0] != '>'  # Markdown quote
        and not line.startswith('#')  # Markdown heading
        and not line.startswith('<!-')  # commented out HTML-style
    ]
    # abbreviate commit hashes in the text
    line = re.sub(r'[a-f0-9]{40}', lambda m: m.group(0)[:7], lines[0])
    # wrap text to get a line of at most 250 chars
    short = textwrap.wrap(line, 250)[0]
    # add continuation marker if needed
    if len(lines) > 1 or short != line:
        short += ' […]'

    return short


def get_distinct_commits(payload=None):
    if not payload:
        payload = current_payload
    if 'distinct_commits' in payload:
        return payload['distinct_commits']
    commits = []
    for commit in payload['commits']:
        if commit['distinct'] and len(commit['message'].strip()) > 0:
            commits.append(commit)
    return commits


def get_ref_name(payload=None):
    if not payload:
        payload = current_payload

    if 'ref_name' in payload:
        return payload['ref_name']

    payload['ref_name'] = re.sub(r'^refs/(heads|tags)/', '', payload['ref'])
    return payload['ref_name']


def get_base_ref_name(payload=None):
    if not payload:
        payload = current_payload
    return re.sub(r'^refs/(heads|tags)/', '', payload['base_ref_name'])


def get_pusher(payload=None):
    if not payload:
        payload = current_payload
    return payload['pusher']['name'] if 'pusher' in payload else 'somebody'


def get_repo_name(payload=None):
    if not payload:
        payload = current_payload
    return payload['repository']['name']


def get_after_sha(payload=None):
    if not payload:
        payload = current_payload
    return payload['after'][0:7]


def get_before_sha(payload=None):
    if not payload:
        payload = current_payload
    return payload['before'][0:7]


def get_push_summary_url(payload=None):
    if not payload:
        payload = current_payload

    repo_url = payload['repository']['url']
    if payload['created'] or re.match(r'0{40}', payload['before']):
        if len(get_distinct_commits()) < 0:
            return repo_url + "/commits/" + get_ref_name()
        else:
            return payload['compare']
    elif payload['deleted']:
        return repo_url + "/commit/" + get_before_sha()
    elif payload['forced']:
        return repo_url + "/commits/" + get_ref_name()
    elif len(get_distinct_commits()) == 1:
        return get_distinct_commits()[0]['url']
    else:
        return payload['compare']


def get_issue_type(payload=None):
    if not payload:
        payload = current_payload

    is_pr = ('pull_request' in payload or ('issue' in payload and '/pull/' in payload['issue']['html_url']))

    if is_pr:
        return "pull request"
    else:
        return "issue"


def get_issue_or_pr_number(payload=None):
    if not payload:
        payload = current_payload

    try:
        number = payload['issue']['number']
    except KeyError:
        number = payload['pull_request']['number']

    return number


def get_issue_or_pr_title(payload=None):
    if not payload:
        payload = current_payload

    try:
        title = payload['issue']['title']
    except KeyError:
        title = payload['pull_request']['title']

    return title


def fmt_push_summary_message(payload=None, row=None):
    if not payload:
        payload = current_payload
    if not row:
        row = current_row

    message = []
    message.append("[{}] {}".format(fmt_repo(get_repo_name()), fmt_name(get_pusher())))

    if payload['created'] or re.match(r'0{40}', payload['before']):
        if re.match(r'^refs/tags/', payload['ref']):
            message.append('tagged {} at'.format(fmt_tag(get_ref_name())))
            message.append(fmt_branch(get_base_ref_name()) if payload['base_ref'] else fmt_hash(get_after_sha()))
        else:
            message.append('created {}'.format(fmt_branch(get_ref_name())))

            if payload['base_ref']:
                message.append('from {}'.format(fmt_branch(get_base_ref_name())))
            elif len(get_distinct_commits()) == 0:
                message.append('at {}'.format(fmt_hash(get_after_sha())))

            num = len(get_distinct_commits())
            message.append('(+\002{}\017 new commit{})'.format(num, 's' if num > 1 else ''))

    elif payload['deleted'] or re.match(r'0{40}', payload['after']):
        message.append("\00304deleted\017 {} at {}".format(fmt_branch(get_ref_name()), fmt_hash(get_before_sha())))

    elif payload['forced']:
        message.append("\00304force-pushed\017 {} from {} to {}".format(
                       fmt_branch(get_ref_name()), fmt_hash(get_before_sha()), fmt_hash(get_after_sha())))

    elif len(payload['commits']) > 0 and len(get_distinct_commits()) == 0:
        if payload['base_ref']:
            message.append('merged {} into {}'.format(fmt_branch(get_base_ref_name()), fmt_branch(get_ref_name())))
        else:
            message.append('fast-forwarded {} from {} to {}'.format(
                           fmt_branch(get_ref_name()), fmt_hash(get_before_sha()), fmt_hash(get_after_sha())))

    else:
        num = len(get_distinct_commits())
        message.append("pushed \002{}\017 new commit{} to {}".format(num, 's' if num > 1 else '', fmt_branch(get_ref_name())))

    return ' '.join(message)


def fmt_commit_message(commit):
    short = commit['message'].splitlines()[0]
    short = short + '…' if short != commit['message'] else short

    author = commit['author']['name']
    sha = commit['id']

    return '{}/{} {} {}: {}'.format(fmt_repo(get_repo_name()), fmt_branch(get_ref_name()), fmt_hash(sha[0:7]), fmt_name(author), short)


def fmt_commit_comment_summary(payload=None, row=None):
    if not payload:
        payload = current_payload
    if not row:
        row = current_row

    short = fmt_short_comment_body(payload['comment']['body'])
    return '[{}] {} commented on commit {}: {}'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  fmt_hash(payload['comment']['commit_id'][0:7]),
                  emojize(short))


def fmt_issue_summary_message(payload=None):
    if not payload:
        payload = current_payload
    return '[{}] {} {} issue #{}: {}'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  payload['action'],
                  payload['issue']['number'],
                  emojize(payload['issue']['title']))


def fmt_issue_incoming_transfer_message(payload=None):
    if not payload:
        payload = current_payload
    # GitHub unfortunately doesn't seem to include any info about the user who
    # initiated the issue transfer, only the original author/creator.
    return '[{}] {}#{} by {} was transferred to issue #{}: {}'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_repo(payload['changes']['old_repository']['full_name']),
                  payload['changes']['old_issue']['number'],
                  fmt_name(payload['issue']['user']['login']),
                  payload['issue']['number'],
                  emojize(payload['issue']['title']))


def fmt_issue_outgoing_transfer_message(payload=None):
    if not payload:
        payload = current_payload
    # For "transferred" events (sent for the source repo only), GitHub DOES set
    # the "sender" info to the user who initiated the transfer, unlike for
    # inbound transfer hooks (which just look like any other "opened" event
    # except for the addition of a "changes" object).
    return '[{}] {} transferred issue #{} by {} to {}#{}: {}'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  payload['issue']['number'],
                  fmt_name(payload['issue']['user']['login']),
                  fmt_repo(payload['changes']['new_repository']['full_name']),
                  payload['changes']['new_issue']['number'],
                  emojize(payload['issue']['title']))


def fmt_issue_title_edit(payload=None):
    if not payload:
        payload = current_payload
    return '[{}] {} retitled issue #{}: "{}" ➜ "{}"'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  payload['issue']['number'],
                  emojize(payload['changes']['title']['from']),
                  emojize(payload['issue']['title']))


def fmt_issue_assignee_message(payload=None):
    if not payload:
        payload = current_payload

    target = ''
    assignee = payload['assignee']['login']
    self_assign = False

    if assignee == payload['sender']['login']:
        self_assign = True
    else:
        prep = 'to' if payload['action'] == 'assigned' else 'from'
        target = ' {} {}'.format(prep, fmt_name(assignee))

    return '[{}] {} {}{} {} #{}{} ({})'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  'self-' if self_assign else '',
                  payload['action'],
                  get_issue_type(payload),
                  get_issue_or_pr_number(payload),
                  target,
                  get_issue_or_pr_title(payload))


def fmt_issue_label_message(payload=None):
    if not payload:
        payload = current_payload
    return '[{}] {} {} the label \'{}\' {} {} #{} ({})'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  'added' if payload['action'] == 'labeled' else 'removed',
                  payload['label']['name'],
                  'to' if payload['action'] == 'labeled' else 'from',
                  get_issue_type(payload),
                  get_issue_or_pr_number(payload),
                  get_issue_or_pr_title(payload))


def fmt_issue_milestone_message(payload=None):
    if not payload:
        payload = current_payload

    added = payload['action'] == 'milestoned'

    return '[{}] {} {} {} #{} ({}) {} the {} milestone'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  'added' if added else 'removed',
                  get_issue_type(payload),
                  get_issue_or_pr_number(payload),
                  get_issue_or_pr_title(payload),
                  'to' if added else 'from',
                  payload['milestone']['title'])


def fmt_issue_comment_summary_message(payload=None):
    if not payload:
        payload = current_payload

    issue_type = get_issue_type(payload)
    short = fmt_short_comment_body(payload['comment']['body'])
    return '[{}] {} commented on {} #{}: {}'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  issue_type,
                  payload['issue']['number'],
                  emojize(short))


def fmt_pull_request_summary_message(payload=None):
    if not payload:
        payload = current_payload

    action = payload['action']
    if action == 'closed' and payload['pull_request']['merged']:
        action = 'merged'
    elif action == 'opened' and payload['pull_request'].get('draft', False):
        action = 'drafted'
    elif action == 'ready_for_review':
        action = 'readied'
    elif action == 'converted_to_draft':
        action = 'un-readied'

    actor = payload['sender']['login']
    author = payload['pull_request']['user']['login']
    maybe_possessive = ''
    if action == 'merged' and actor != author:
        maybe_possessive = '%s\'s ' % fmt_name(author)

    base = fmt_branch(payload['pull_request']['base']['ref'])
    head = fmt_branch(payload['pull_request']['head']['ref'])
    base_repo = payload['pull_request']['base']['user']['login']
    head_repo = payload['pull_request']['head']['user']['login']
    if base_repo != head_repo:
        base = "{}:{}".format(fmt_name(base_repo), base)
        head = "{}:{}".format(fmt_name(head_repo), head)

    return '[{}] {} {} {}pull request #{}: {} ({}...{})'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(actor),
                  action,
                  maybe_possessive,
                  payload['pull_request']['number'],
                  emojize(payload['pull_request']['title']),
                  base,
                  head)


def fmt_pull_request_title_edit(payload=None):
    if not payload:
        payload = current_payload
    return '[{}] {} retitled PR #{}: "{}" ➜ "{}"'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  payload['pull_request']['number'],
                  emojize(payload['changes']['title']['from']),
                  emojize(payload['pull_request']['title']))


def fmt_pull_request_review_summary_message(payload=None):
    if not payload:
        payload = current_payload

    action = payload['review']['state']
    if action == 'commented':
        action = 'left a review on'
    elif action == 'changes_requested':
        action = 'requested changes on'

    body = payload['review']['body']
    short = ''
    if body:
        short = fmt_short_comment_body(body)
        short = ': ' + short

    return '[{}] {} {} pull request #{}{}'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  action,
                  payload['pull_request']['number'],
                  emojize(short))


def fmt_pull_request_review_dismissal_message(payload=None):
    if not payload:
        payload = current_payload

    if payload['sender']['login'] == payload['review']['user']['login']:
        whose = 'their'
    else:
        whose = fmt_name(payload['review']['user']['login']) + '\'s'

    return '[{}] {} dismissed {} review on pull request #{}'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  whose,
                  payload['pull_request']['number'])


def fmt_pull_request_review_comment_summary_message(payload=None):
    if not payload:
        payload = current_payload
    short = fmt_short_comment_body(payload['comment']['body'])
    sha1 = payload['comment']['commit_id']
    return '[{}] {} left a file comment in pull request #{} {}: {}'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  payload['pull_request']['number'],
                  fmt_hash(sha1[0:7]),
                  emojize(short))


def fmt_gollum_summary_message(payload=None):
    if not payload:
        payload = current_payload
    if len(payload['pages']) == 1:
        summary = None
        if 'summary' in payload['pages'][0]:
            summary = payload['pages'][0]['summary']

        return '[{}] {} {} wiki page {}{}'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  payload['pages'][0]['action'],
                  payload['pages'][0]['title'],
                  ": " + summary if summary else '')
    elif len(payload['pages']) > 1:
        counts = {}
        for page in payload['pages']:
            # Set default value to 0 and increment 1, only incrementing if key already exists
            counts[page['action']] = counts.setdefault(page['action'], 0) + 1
        actions = []
        for action, count in counts.items():
            actions.append(action + " " + count)

        return '[{}] {} {} wiki pages'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']),
                  fmt_arr_to_sentence(actions.sort()))


def fmt_arr_to_sentence(array):
    if len(seq) <= 2:
        return ' and '.join(seq)
    else:
        return '{}, and {}'.format(', '.join(seq[:-1]), seq[-1])


def fmt_watch_message(payload=None):
    if not payload:
        payload = current_payload
    return '[{}] {} starred the project!'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_name(payload['sender']['login']))


def fmt_status_message(payload=None):
    if not payload:
        payload = current_payload
    branch = ''
    for br in payload['branches']:
        if br['commit']['sha'] == payload['sha']:
            branch = br['name']
    return '[{}/{}] {} - {} ({})'.format(
                  fmt_repo(payload['repository']['name']),
                  fmt_branch(branch),
                  payload['description'],
                  payload['target_url'],
                  payload['state'])


def fmt_release_message(payload=None):
    if not payload:
        payload = current_payload

    return '[{}] {} released {}{}'.format(
        fmt_repo(payload['repository']['name']),
        fmt_name(payload['release']['author']['login']),
        payload['release']['name'] or payload['release']['tag_name'],
        ' (prerelease)' if payload['release']['prerelease'] else '')


def shorten_url(url):
    try:
        res = requests.post('https://git.io', {'url': url})
        return res.headers['location']
    except:
        LOGGER.exception('Shortening link failed; using long URL.')
        return url


def get_formatted_response(payload, row):
    global current_row, current_payload
    current_payload = payload
    current_row = row

    messages = []
    if payload['event'] == 'push':
        messages.append(fmt_push_summary_message() + " " + fmt_url(shorten_url(get_push_summary_url())))
        for commit in get_distinct_commits():
            messages.append(fmt_commit_message(commit))
    elif payload['event'] == 'commit_comment':
        messages.append(fmt_commit_comment_summary() + " " + fmt_url(shorten_url(payload['comment']['html_url'])))
    elif payload['event'] == 'pull_request':
        if re.match('((re)?open|clos)ed', payload['action']) or payload['action'] in ['ready_for_review', 'converted_to_draft']:
            messages.append(fmt_pull_request_summary_message() + " " + fmt_url(shorten_url(payload['pull_request']['html_url'])))
        elif payload['action'] == 'edited':
            if 'changes' in payload:
                if 'title' in payload['changes']:
                    messages.append(fmt_pull_request_title_edit() + " " + fmt_url(shorten_url(payload['pull_request']['html_url'])))
        elif re.match('(assigned|unassigned)', payload['action']):
            messages.append(fmt_issue_assignee_message() + " " + fmt_url(shorten_url(payload['pull_request']['html_url'])))
        elif re.match('(labeled|unlabeled)', payload['action']):
            if payload.get('label', None):
                # If a label is deleted, for example, we'll get a webhook payload with no details about the removed label.
                # We skip those; there's no reason to emit action messages to IRC with "unknown label" placeholders.
                messages.append(fmt_issue_label_message() + " " + fmt_url(shorten_url(payload['pull_request']['html_url'])))
    elif payload['event'] == 'pull_request_review':
        if payload['action'] == 'submitted' and payload['review']['state'] in ['approved', 'changes_requested', 'commented']:
            if payload['review']['state'] == 'commented' and payload['review']['body'] == None:
                # Probably an empty "review" fired by a pull_request_review_comment reply, which we'll get in a separate hook delivery.
                # Wish GitHub didn't fire both events, but they do, even though it makes no sense.
                # Either way, an empty review must be accompanied by comments, which will get handled when their hook(s) fire(s).
                pass
            else:
                messages.append(fmt_pull_request_review_summary_message() + " " + fmt_url(shorten_url(payload['review']['html_url'])))
        elif payload['action'] == 'dismissed':
            messages.append(fmt_pull_request_review_dismissal_message() + " " + fmt_url(shorten_url(payload['review']['html_url'])))
    elif payload['event'] == 'pull_request_review_comment' and payload['action'] == 'created':
        messages.append(fmt_pull_request_review_comment_summary_message() + " " + fmt_url(shorten_url(payload['comment']['html_url'])))
    elif payload['event'] == 'issues':
        if re.match('((re)?open|clos)ed', payload['action']):
            if 'changes' in payload and all(k in payload['changes'] for k in ['old_repository', 'old_issue']):
                messages.append(fmt_issue_incoming_transfer_message() + " " + fmt_url(shorten_url(payload['issue']['html_url'])))
            else:
                messages.append(fmt_issue_summary_message() + " " + fmt_url(shorten_url(payload['issue']['html_url'])))
        elif re.match('(assigned|unassigned)', payload['action']):
            messages.append(fmt_issue_assignee_message() + " " + fmt_url(shorten_url(payload['issue']['html_url'])))
        elif re.match('(labeled|unlabeled)', payload['action']):
            if payload.get('label', None):
                # If a label is deleted, for example, we'll get a webhook payload with no details about the removed label.
                # We skip those; there's no reason to emit action messages to IRC with "unknown label" placeholders.
                messages.append(fmt_issue_label_message() + " " + fmt_url(shorten_url(payload['issue']['html_url'])))
        elif re.match('(milestoned|demilestoned)', payload['action']):
            messages.append(fmt_issue_milestone_message() + " " + fmt_url(shorten_url(payload['issue']['html_url'])))
        elif payload['action'] == 'edited':
            if 'changes' in payload:
                if 'title' in payload['changes']:
                    messages.append(fmt_issue_title_edit() + " " + fmt_url(shorten_url(payload['issue']['html_url'])))
        elif payload['action'] == 'transferred':
            messages.append(fmt_issue_outgoing_transfer_message() + " " + fmt_url(shorten_url(payload['changes']['new_issue']['html_url'])))
    elif payload['event'] == 'issue_comment' and payload['action'] == 'created':
        messages.append(fmt_issue_comment_summary_message() + " " + fmt_url(shorten_url(payload['comment']['html_url'])))
    elif payload['event'] == 'gollum':
        url = payload['pages'][0]['html_url'] if len(payload['pages']) else payload['repository']['url'] + '/wiki'
        messages.append(fmt_gollum_summary_message() + " " + fmt_url(shorten_url(url)))
    elif payload['event'] == 'watch':
        messages.append(fmt_watch_message())
    elif payload['event'] == 'status':
        messages.append(fmt_status_message())
    elif payload['event'] == 'release':
        if payload['action'] == 'published':
            # Currently the only possible action, but other events might eventually fire webhooks too
            messages.append(fmt_release_message() + " " + fmt_url(shorten_url(payload['release']['html_url'])))

    return messages
